#!/usr/bin/env python

import optparse, os, shutil, subprocess, sys, tempfile

#Allow more than Sanger encoded variants
DEFAULT_ASCII_ENCODING = '--phred33-quals'
GALAXY_FORMAT_TO_QUALITY_SCORE_ENCODING_ARG = { 'fastqsanger':'--phred33-quals', 'fastqillumina':'--phred64-quals', 'fastqsolexa':'--solexa-quals' }
#FIXME: Integer quality scores are supported only when the '--integer-quals' argument is specified to bowtie; this is not currently able to be set in the tool/wrapper/config

def stop_err( msg ):
    sys.stderr.write( '%s\n' % msg )
    sys.exit()

def __main__():
    #Parse Command Line
    parser = optparse.OptionParser()
    parser.add_option( '-t', '--threads', dest='threads', help='The number of threads to run' )
    parser.add_option( '-n', '-n', dest='validal', help='The n param' )
    parser.add_option( '-m', '-m', dest='suppal', help='The m param' )
    parser.add_option( '-k', '-k', dest='reportvalal', help='The k param' )
    parser.add_option( '-l', '-l', dest='seedlength', help='The l param' )
    parser.add_option( '-e', '-e', dest='maqerr', help='The e param' )
    parser.add_option( '-4', '--dataType', dest='dataType', help='The type of data (SOLiD or Solexa)' )
    parser.add_option( '-g', '--genomeSource', dest='genomeSource', help='The type of reference provided' )
    parser.add_option( '-r', '--ref', dest='ref', help='The reference genome to use or index' )
    parser.add_option( '--do_not_build_index', dest='do_not_build_index', action="store_true", default=False, help='Flag to specify that provided file is already indexed, use as is' )
    parser.add_option( '-x', '--indexSettings', dest='index_settings', help='Whether or not indexing options are to be set' )
    parser.add_option( '-u', '--iautoB', dest='iautoB', help='Automatic or specified behavior' )
    parser.add_option( '-K', '--ipacked', dest='ipacked', help='Whether or not to use a packed representation for DNA strings' )
    parser.add_option( '-Q', '--ibmax', dest='ibmax', help='Maximum number of suffixes allowed in a block' )
    parser.add_option( '-Y', '--ibmaxdivn', dest='ibmaxdivn', help='Maximum number of suffixes allowed in a block as a fraction of the length of the reference' )
    parser.add_option( '-D', '--idcv', dest='idcv', help='The period for the difference-cover sample' )
    parser.add_option( '-U', '--inodc', dest='inodc', help='Whether or not to disable the use of the difference-cover sample' )
    parser.add_option( '-y', '--inoref', dest='inoref', help='Whether or not to build the part of the reference index used only in paired-end alignment' )
    parser.add_option( '-z', '--ioffrate', dest='ioffrate', help='How many rows get marked during annotation of some or all of the Burrows-Wheeler rows' )
    parser.add_option( '-W', '--iftab', dest='iftab', help='The size of the lookup table used to calculate an initial Burrows-Wheeler range with respect to the first n characters of the query' )
    parser.add_option( '-X', '--intoa', dest='intoa', help='Whether or not to convert Ns in the reference sequence to As' )
    parser.add_option( '-N', '--iendian', dest='iendian', help='Endianness to use when serializing integers to the index file' )
    parser.add_option( '-Z', '--iseed', dest='iseed', help='Seed for the pseudorandom number generator' )
    parser.add_option( '-c', '--icutoff', dest='icutoff', help='Number of first bases of the reference') 
    parser.add_option( '-f', '-f', dest='input', help='The input file' )
    parser.add_option( '-S', '-S', dest='samformat', help='The output format' )
    parser.add_option( '-o', '-o', dest='output', help='The output file' )
    parser.add_option( '--galaxy_input_format', dest='galaxy_input_format', default="fasta", help='galaxy input format' )
    (options, args) = parser.parse_args()
    stdout = ''
    
    # make temp directory for placement of indices and copy reference file there if necessary
    tmp_index_dir = tempfile.mkdtemp()
    # get type of data (solid or solexa)
    if options.dataType == 'solid':
        colorspace = '-C'
    else:
        colorspace = ''
	
	    # get type of data (solid or solexa)
    if options.samformat == 'true':
        format = '-S'
    else:
        format = ''	
		
    # index if necessary
    if options.genomeSource == 'history' and not options.do_not_build_index:
        # set up commands
        if options.index_settings =='indexPreSet':
            indexing_cmds = '%s' % colorspace
        else:
            try:
                if options.iautoB and options.iautoB == 'set':
                    iautoB = '--noauto'
                else:
                    iautoB = ''
                if options. ipacked and options.ipacked == 'packed':
                    ipacked = '--packed'
                else:
                    ipacked = ''
                if options.ibmax and int( options.ibmax ) >= 1:
                    ibmax = '--bmax %s' % options.ibmax 
                else:
                    ibmax = ''
                if options.ibmaxdivn and int( options.ibmaxdivn ) >= 0:
                    ibmaxdivn = '--bmaxdivn %s' % options.ibmaxdivn
                else:
                    ibmaxdivn = ''
                if options.idcv and int( options.idcv ) > 0:
                    idcv = '--dcv %s' % options.idcv
                else:
                    idcv = ''
                if options.inodc and options.inodc == 'nodc':
                    inodc = '--nodc'
                else:
                    inodc = ''
                if options.inoref and options.inoref == 'noref':
                    inoref = '--noref'
                else:
                    inoref = ''
                if options.iftab and int( options.iftab ) >= 0:
                    iftab = '--ftabchars %s' % options.iftab
                else:
                    iftab = ''
                if options.intoa and options.intoa == 'yes':
                    intoa = '--ntoa'
                else:
                    intoa = ''
                if options.iendian and options.iendian == 'big':
                    iendian = '--big'
                else:
                    iendian = '--little'
                if options.iseed and int( options.iseed ) > 0:
                    iseed = '--seed %s' % options.iseed
                else:
                    iseed = ''
                if options.icutoff and int( options.icutoff ) > 0:
                    icutoff = '--cutoff %s' % options.icutoff
                else:
                    icutoff = ''
                indexing_cmds = '%s %s %s %s %s %s %s --offrate %s %s %s %s %s %s %s' % \
                                ( iautoB, ipacked, ibmax, ibmaxdivn, idcv, inodc, 
                                  inoref, options.ioffrate, iftab, intoa, iendian, 
                                  iseed, icutoff, colorspace )
            except ValueError, e:
                # clean up temp dir
                if os.path.exists( tmp_index_dir ):
                    shutil.rmtree( tmp_index_dir )
                stop_err( "Something is wrong with the indexing parameters and the indexing and alignment could not be run. Make sure you don't have any non-numeric values where they should be numeric.\n" + str( e ) )
        ref_file = tempfile.NamedTemporaryFile( dir=tmp_index_dir )
        ref_file_name = ref_file.name
        ref_file.close()
        os.symlink( options.ref, ref_file_name )

        cmd1 = 'bowtie-build %s -f %s %s' % ( indexing_cmds, ref_file_name, ref_file_name )
	
        try:
            tmp = tempfile.NamedTemporaryFile( dir=tmp_index_dir ).name
            tmp_stderr = open( tmp, 'wb' )
            proc = subprocess.Popen( args=cmd1, shell=True, cwd=tmp_index_dir, stderr=tmp_stderr.fileno() )
            returncode = proc.wait()
            tmp_stderr.close()
            # get stderr, allowing for case where it's very large
            tmp_stderr = open( tmp, 'rb' )
            stderr = ''
            buffsize = 1048576
            try:
                while True:
                    stderr += tmp_stderr.read( buffsize )
                    if not stderr or len( stderr ) % buffsize != 0:
                        break
            except OverflowError:
                pass
            tmp_stderr.close()
            if returncode != 0:
                raise Exception, stderr
        except Exception, e:
            # clean up temp dir
            if os.path.exists( tmp_index_dir ):
                shutil.rmtree( tmp_index_dir )
            stop_err( 'Error indexing reference sequence\n' + str( e ) )
        stdout += 'File indexed. '
    else:

        ref_file_name = options.ref
    # set up aligning and generate aligning command options
    # automatically set threads in both cases
    tmp_suppressed_file_name = None
    tmp_unmapped_file_name = None
    quality_score_encoding = GALAXY_FORMAT_TO_QUALITY_SCORE_ENCODING_ARG.get( options.galaxy_input_format, DEFAULT_ASCII_ENCODING )
    try:
        # have to nest try-except in try-finally to handle 2.4
        try:
            # prepare actual mapping commands
            cmd2 = 'bowtie -n %s -m %s -k %s -l %s -e %s %s -f %s %s > %s' % ( options.validal, options.suppal, options.reportvalal, options.seedlength, options.maqerr,options.ref, options.input, format, options.output)
            print(cmd2)
            # align
            tmp = tempfile.NamedTemporaryFile( dir=tmp_index_dir ).name
            tmp_stderr = open( tmp, 'wb' )
            proc = subprocess.Popen( args=cmd2, shell=True, cwd=tmp_index_dir, stderr=tmp_stderr.fileno() )
            returncode = proc.wait()
            tmp_stderr.close()
            # get stderr, allowing for case where it's very large
            tmp_stderr = open( tmp, 'rb' )
            stderr = ''
            buffsize = 1048576
            try:
                while True:
                    stderr += tmp_stderr.read( buffsize )
                    if not stderr or len( stderr ) % buffsize != 0:
                        break
            except OverflowError:
                pass
            tmp_stderr.close()
            if returncode != 0:
                raise Exception, stderr
            # check that there are results in the output file
            if os.path.getsize( options.output ) == 0:
                raise Exception, 'The output file is empty, there may be an error with your input file or settings.'
        except Exception, e:
            stop_err( 'Error aligning sequence. ' + str( e ) )
    finally:
        # clean up temp dir
        if os.path.exists( tmp_index_dir ):
            shutil.rmtree( tmp_index_dir )
    stdout += 'Sequence file aligned.\n'
    sys.stdout.write( stdout )

if __name__=="__main__": __main__()
